
#ifndef _GDK_H_
#define _GDK_H_

#include <monet_utils.h>

/* standard includes upon which all configure tests depend */
#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#ifdef HAVE_INTTYPES_H
# include <inttypes.h>
#else
# ifdef HAVE_STDINT_H
#  include <stdint.h>
# endif
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <ctype.h>		/* isspace etc. */

#ifdef HAVE_SYS_FILE_H
# include <sys/file.h>
#endif
#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>		/* MAXPATHLEN */
#endif

#ifdef HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif


#ifdef __cplusplus
extern "C" {
#endif

#include "gdk_system.h"
#include "gdk_posix.h"
#include <stream.h>

#undef MIN
#undef MAX
#define MAX(A,B)	((A)<(B)?(B):(A))
#define MIN(A,B)	((A)>(B)?(B):(A))

/* defines from ctype with casts that allow passing char values */
#define GDKisprint(c)   isprint((int) ((unsigned char) (c)))
#define GDKisspace(c)   isspace((int) ((unsigned char) (c)))
#define GDKisalnum(c)   isalnum((int) ((unsigned char) (c)))
#define GDKisgraph(c)   isgraph((int) ((unsigned char) (c)))
#define GDKisdigit(c)   (((unsigned char) (c)) >= '0' && ((unsigned char) (c)) <= '9')
#define GDKisxcntrl(c)  (((unsigned char) (c)) >= 128 && ((unsigned char) (c)) <= 160)
#define GDKisspecial(c) (((unsigned char) (c)) >= 161 && ((unsigned char) (c)) <= 191)
#define GDKisupperl(c)  (((unsigned char) (c)) >= 192 && ((unsigned char) (c)) <= 223)
#define GDKislowerl(c)  (((unsigned char) (c)) >= 224 && ((unsigned char) (c)) <= 255)

#define GDKPROP		6	/*      use one spare! */
#define MONETHOME	"MONETHOME"
#ifndef NATIVE_WIN32
#define BATDIR          "bat"
#define DELDIR          "bat/DELETE_ME"
#define BAKDIR          "bat/BACKUP"
#define SUBDIR          "bat/BACKUP/SUBCOMMIT"
#define LEFTDIR         "bat/LEFTOVERS"
#else
#define BATDIR          "bat"
#define DELDIR          "bat\\DELETE_ME"
#define BAKDIR          "bat\\BACKUP"
#define SUBDIR          "bat\\BACKUP\\SUBCOMMIT"
#define LEFTDIR         "bat\\LEFTOVERS"
#endif
#define DBLOGFILE	"MonetLog"

#ifdef MAXPATHLEN
#define PATHLENGTH 	MAXPATHLEN
#else
#define PATHLENGTH	1024	/* maximum file pathname length */
#endif



#define NEG(A)	(((int)(A))>0?-((int)(A)):((int)(A)))
#define ABS(A)	(((int)(A))>0?((int)(A)):-((int)(A)))

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif
#define BOUND2BTRUE	2	/* TRUE, and bound to be so */

#define IDLENGTH	64	/* maximum BAT id length */
#define OLDLENGTH	32	/* old max  BAT id length */
#define BATMARGIN	1.2	/* extra free margin for new heaps */
#define BATTINY_BITS 	8
#define BATTINY		(1<<BATTINY_BITS)	/* minimum allocation buncnt for a BAT */

#define TYPE_void	0
#define TYPE_bit	1
#define TYPE_chr	2
#define TYPE_bte	3
#define TYPE_sht	4
#define TYPE_bat	5	/* BAT id: index in BBPcache */
#define TYPE_int	6
#define TYPE_oid	7
#define TYPE_wrd	8
#define TYPE_ptr	9	/* C pointer! */
#define TYPE_flt	10
#define TYPE_dbl	11
#define TYPE_lng	12
#define TYPE_str	13
#define TYPE_any	255	/* limit types to <255! */

typedef signed char bit;
typedef signed char chr;
typedef signed char bte;
typedef short sht;

#ifdef MONET_OID32
#define SIZEOF_OID	SIZEOF_INT
typedef unsigned int oid;
#else
#define SIZEOF_OID	SIZEOF_SIZE_T
typedef size_t oid;
#endif
#define SIZEOF_WRD      SIZEOF_SSIZE_T
typedef ssize_t wrd;
typedef int bat;		/* Index into BBP */
typedef void *ptr;		/* Internal coding of types */

#define SIZEOF_PTR	SIZEOF_VOID_P
typedef float flt;
typedef double dbl;
typedef char *str;

#if SIZEOF_INT==8
#	define LL_CONSTANT(val)     (val)
#elif SIZEOF_LONG==8
#	define LL_CONSTANT(val)     (val##L)
#elif defined(HAVE_LONG_LONG)
#	define LL_CONSTANT(val)     (val##LL)
#elif defined(HAVE___INT64)
#	define LL_CONSTANT(val)     (val##i64)
#endif

/* Base types for all architectures.  These types work on IRIX (64 & 32)
   and for gcc on Linux. */
typedef unsigned char gdk_uint8;
typedef char gdk_int8;

typedef unsigned short gdk_uint16;
typedef short gdk_int16;

typedef unsigned int gdk_uint32;
typedef int gdk_int32;

#ifdef HAVE_LONG_LONG
typedef unsigned long long gdk_uint64;
typedef long long gdk_int64;
#else
#ifdef HAVE___INT64
typedef unsigned __int64 gdk_uint64;
typedef __int64 gdk_int64;
#endif
#endif

/* define printf formats for printing ptrdiff_t variables */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901 && !defined(__svr4__) && !defined(WIN32) && !defined(__sgi)
#define PDFMT "%td"
#elif SIZEOF_PTRDIFF_T == SIZEOF_INT
#define PDFMT "%d"
#elif SIZEOF_PTRDIFF_T == SIZEOF_LONG
#define PDFMT "%ld"
#elif SIZEOF_PTRDIFF_T == SIZEOF_LONG_LONG || SIZEOF_PTRDIFF_T == SIZEOF___INT64
#define PDFMT LLFMT
#else
#error no definition for PDFMT
#endif

/* define printf format for printing pointer values */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901
#define PTRFMT		"%p"
#define PTRFMTCAST		/* no cast needed */
#elif SIZEOF_VOID_P == SIZEOF_INT
#define PTRFMT		"%x"
#define PTRFMTCAST	(unsigned int)
#elif SIZEOF_VOID_P == SIZEOF_LONG
#define PTRFMT		"%lx"
#define PTRFMTCAST	(unsigned long)
#elif SIZEOF_VOID_P == SIZEOF_LONG_LONG
#define PTRFMT		"%llx"
#define PTRFMTCAST	(unsigned long long)
#elif SIZEOF_VOID_P == SIZEOF___INT64
#define PTRFMT		"%I64x"
#define PTRFMTCAST	(unsigned __int64)
#else
#error no definition for PTRFMT
#endif

typedef oid BUN;		/* BUN position */
typedef char long_str[IDLENGTH];	/* standard GDK static string */
typedef char short_str[OLDLENGTH];	/* old GDK small static string */

#if SIZEOF_OID == SIZEOF_INT
#define BUN_NONE INT_MAX
#else
#define BUN_NONE LLONG_MAX
#endif

typedef oid hash_t;		/* type used for hash tables */

#if SIZEOF_OID == SIZEOF_INT
#define HASH_MAX INT_MAX
#else
#define HASH_MAX LLONG_MAX
#endif

typedef oid var_t;		/* type used for heap index of var-sized BAT */

#if SIZEOF_OID == SIZEOF_INT	/* a type compatible with var_t */
#define TYPE_var	TYPE_int
#if SIZEOF_SIZE_T == SIZEOF_INT
#define OIDFMT		SZFMT
#else
#define OIDFMT		"%u"
#endif
#else
#define TYPE_var	TYPE_lng
#define OIDFMT		SZFMT
#endif



typedef enum { GDK_FAIL, GDK_SUCCEED } gdk_return;

gdk_export int GDKsilent;	/* should GDK shut up? */

#define FATALcheck(tst,	msg) if (tst) GDKfatal(msg);
#define ERRORcheck(tst,	msg) if (tst) { GDKerror(msg); return 0; }
#define WARNcheck(tst,	msg) if (tst) GDKwarning(msg)
#define BATcheck(tst,	msg) if (tst == NULL) {				\
				GDKerror("%s: BAT required.\n",msg);	\
				return 0;				\
			}

/* needed for mel */
#define ATOMextern(t)	(ATOMstorage(t) >= TYPE_str)

#define TYPEcastable(t1,t2)	(ATOMtype(t1)==ATOMtype(t2))
#define TYPEequal(t1,t2)	(ATOMtype(t1)==ATOMtype(t2))
#define TYPEcomp(t1,t2)	(ATOMstorage(ATOMtype(t1))==ATOMstorage(ATOMtype(t2)))
#define TYPEerror(t1,t2)	(!TYPEcomp(t1,t2))
#define TYPEcheck(t1,t2)						\
	if (TYPEerror(t1,t2)) {						\
		GDKerror("TYPEcheck: Incompatible types %s and %s.\n",	\
			ATOMname(t2), ATOMname(t1));			\
		return 0;						\
	} else if (!TYPEcomp(t1,t2)) {					\
		GDKwarning("Interpreting %s as %s.\n",			\
			ATOMname(t2), ATOMname(t1));			\
	}
#define BATcompatible(P1,P2) {						\
	ERRORcheck(P1 == NULL, "BATcompatible: BAT required\n");	\
	ERRORcheck(P2 == NULL, "BATcompatible: BAT required\n");	\
	if (TYPEerror(BAThtype(P1),BAThtype(P2)) ||			\
	    TYPEerror(BATttype(P1),BATttype(P2)))			\
	{								\
		GDKerror("Incompatible operands.\n");			\
		return 0;						\
	}								\
	if (BAThtype(P1) != BAThtype(P2) &&                             \
	    ATOMtype((P1)->htype) != ATOMtype((P2)->htype)) {		\
		GDKwarning("Interpreting %s as %s.\n",			\
			ATOMname(BAThtype(P2)), ATOMname(BAThtype(P1))); \
	}								\
	if (BATttype(P1) != BATttype(P2) &&                             \
	    ATOMtype((P1)->ttype) != ATOMtype((P2)->ttype)) {		\
		GDKwarning("Interpreting %s as %s.\n",			\
			ATOMname(BATttype(P2)), ATOMname(BATttype(P1))); \
	}								\
}


typedef struct {
	size_t maxsize;		/* maximum realloc size (bytes) */
	size_t free;		/* index where free area starts. */
	size_t size;		/* size of the heap (bytes) */
	char *base;		/* base pointer in memory. */
	str filename;		/* file containing image of the heap */

	char storage;		/* storage mode (mmap/malloc). */
	bit copied;		/* a copy of an existing map. */
	chr newstorage;		/* new desired storage mode at re-allocation. */
	chr dirty;		/* specific heap dirty marker */
} Heap;

typedef struct {
	int type;		/* type of index entity */
	size_t lim;		/* collision list size */
	hash_t mask;		/* number of hash buckets-1 (power of 2) */
	hash_t *hash;		/* hash table */
	hash_t *link;		/* collision list */
	Heap *heap;		/* heap where the hash is stored */
} Hash;



typedef struct {
	bat id;
	int stamp;
} batrec;

typedef struct {
	union {			/* storage is first in the record */
		int ival;
		oid oval;
		sht shval;
		bte btval;
		wrd wval;
		chr cval[4];
		flt fval;
		ptr pval;
		struct BAT *Bval;	/* this field is only used by mel */
		bat bval;
		batrec br;	/* bval + stamp */
		str sval;
		dbl dval;
		lng lval;
	} val;
	int len, vtype;
} *ValPtr, ValRecord;

#define VALptr(v) (ATOMextern((v)->vtype)?(v)->val.pval:(ptr)&(v)->val.ival)
#define VALnil(v,t) VALset(v,t,ATOMextern(t)?ATOMnil(t):ATOMnilptr(t))

/* interface definitions */
gdk_export ValPtr VALnew(void);
gdk_export ptr VALconvert(int typ, ValPtr t);
gdk_export ptr VALconvert1(int typ, ValPtr src, ValPtr dst);
gdk_export ptr VALconvert2(int typ, ValPtr src, ValPtr dst);
gdk_export int VALformat(char **buf, ValPtr res);
gdk_export int VALprint(stream *fd, ValPtr res);
gdk_export ValPtr VALcopy(ValPtr dst, ValPtr src);
gdk_export ValPtr VALinit(ValPtr d, int tpe, ptr s);
gdk_export void VALempty(ValPtr v);
gdk_export void VALclear(ValPtr v);
gdk_export ValPtr VALset(ValPtr v, int t, ptr p);
gdk_export void *VALget(ValPtr v);







typedef struct {
	MT_Id tid;		/* which thread created it */
	int stamp;		/* BAT recent creation stamp */
	unsigned int
	 copiedtodisk:1,	/* once written */
	 dirty:2,		/* dirty wrt disk? */
	 dirtyflushed:1,	/* was dirty before commt started? */
	 descdirty:1,		/* bat descriptor dirty marker */
	 lview:1,		/* bat is a *logical* view on parentid */
	 unused:26;		/* value=0 for now */
	chr set;		/* real set semantics */
	chr restricted;		/* access priviliges */
	sht persistence;	/* should the BAT persist on disk? */
	int sharecnt;		/* incoming view count */
	bit map_dirty;		/* new mmap mode present */
	char map_buns;		/* mmap mode for bun heap */
	char map_hheap;		/* mmap mode for head atom heap */
	char map_theap;		/* mmap mode for tail atom heap */
} BATrec;

typedef struct {
	/* delta status administration */
	BUN deleted;		/* start of deleted elements */
	BUN first;		/* to store next deletion */
	BUN inserted;		/* start of inserted elements */
	size_t count;		/* tuple count */
	size_t capacity;	/* tuple capacity */
} BUNrec;

typedef struct PROPrec {
	int id;
	ValRecord v;
	struct PROPrec *next;	/* simple chain of properties */
} PROPrec;

typedef struct {
	str id;			/* label for head/tail column */

	unsigned short width;	/* byte-width of the atom array */
	bit varsized;		/* varsized(>0) or fixedsized(0). */
	chr type;		/* type id. */
	chr shift;		/* log2 of bunwidth */
	bit key;		/* duplicates allowed? */
	bit sorted;		/* 0=false, 1=true; */
	bit dense;		/* 0=false, 1=true; */
	oid align;		/* OID for sync alignment */
	size_t nosorted_rev;	/* position that proves sorted_rev==FALSE */
	size_t nokey[2];	/* positions that prove key ==FALSE */
	size_t nosorted;	/* position that proves sorted==FALSE */
	size_t nodense;		/* position that proves dense==FALSE */
	oid seq;		/* start of dense head sequence */

	Heap heap;		/* space for the column. */
	Heap *vheap;		/* space for the varsized data. */
	Hash *hash;		/* hash table */

	PROPrec *props;		/* list of dynamic properties stored in the bat descriptor */
	bat parentid;		/* cache id of VIEW parent bat */
} COLrec;

#define GDKLIBRARY      061022	/* 64-bits BAT */

typedef struct BAT {
	/* static bat properties */
	sht GDKversion;		/* FIRST! */
	bat batCacheid;		/* index into BBP */

	/* dynamic column properties */
	COLrec *H;		/* column info */
	COLrec *T;		/* column info */

	/* dynamic bat properties */
	BATrec *P;		/* cache and sort info */
	BUNrec *U;		/* cache and sort info */
} BAT;

typedef struct BATiter {
	BAT *b;
	oid hvid, tvid;
} BATiter;



typedef struct {
	BAT B;			/* storage for BAT descriptor */
	COLrec H;		/* storage for head column */
	COLrec T;		/* storage for tail column */
	BATrec P;		/* storage for BATrec */
	BUNrec U;		/* storage for BUNrec */
} BATstore;

typedef int (*GDKfcn) ();

/* macros's to hide complexity of BAT structure */
#define batPersistence	P->persistence
#define batCopiedtodisk	P->copiedtodisk
#define batSet		P->set
#define batDirtyflushed	P->dirtyflushed
#define batDirty	P->dirty
#define batConvert	P->convert
#define batDirtyflushed P->dirtyflushed
#define batDirtydesc	P->descdirty
#define batFirst	U->first
#define batInserted	U->inserted
#define batDeleted	U->deleted
#define batCount	U->count
#define batCapacity	U->capacity
#define batStamp	P->stamp
#define batSharecnt	P->sharecnt
#define batRestricted	P->restricted
#define batLastused	P->lastused
#define creator_tid	P->tid
#define htype		H->type
#define ttype		T->type
#define hkey		H->key
#define tkey		T->key
#define hvarsized	H->varsized
#define tvarsized	T->varsized
#define hseqbase	H->seq
#define tseqbase	T->seq
#define hsorted		H->sorted
#define tsorted		T->sorted
#define hdense		H->dense
#define tdense		T->dense
#define hheap		H->vheap
#define theap		T->vheap
#define hident		H->id
#define tident		T->id
#define halign		H->align
#define talign		T->align

#define batMapdirty	P->map_dirty
#define batMapbuns	P->map_buns
#define batMaphheap	P->map_hheap
#define batMaptheap	P->map_theap


gdk_export int HEAPalloc(Heap *h, size_t nitems, size_t itemsize);
gdk_export int HEAPfree(Heap *h);
gdk_export int HEAPextend(Heap *h, size_t size);
gdk_export int HEAPload(Heap *h, str nme, str ext, int trunc);
gdk_export int HEAPsave(Heap *h, str nme, str ext);
gdk_export int HEAPcopy(Heap *dst, Heap *src);
gdk_export int HEAPwarm(Heap *h);
gdk_export size_t HEAPsubcopy(Heap *dst, Heap *src, size_t size, size_t offset);
gdk_export int HEAPdelete(Heap *h, str o, str ext);
gdk_export size_t HEAPvmsize(Heap *h);
gdk_export size_t HEAPmemsize(Heap *h);





/* structure used by HEAP_check functions */
typedef struct {
	size_t minpos;		/* minimum block byte-index */
	size_t maxpos;		/* maximum block byte-index */
	int alignment;		/* block index alignment */
	size_t *validmask;	/* bitmap with all valid byte-indices
				 * first bit corresponds with 'minpos';
				 * 2nd bit with 'minpos+alignment', etc
				 */
} HeapRepair;

gdk_export void HEAP_initialize(Heap *heap,	/* nbytes -- Initial size of the heap. */
				size_t nbytes,	/* alignment -- for objects on the heap. */
				size_t nprivate,/* nprivate -- Size of private space */
				int alignment	/* alignment restriction for allocated chunks */
    );

gdk_export void HEAP_initialize_compact(Heap *heap,	/* nbytes -- Initial size of the heap. */
					size_t nbytes,	/* alignment -- for objects on the heap. */
					size_t nprivate,/* nprivate -- Size of private space */
					int alignment,	/* alignment restriction for allocated chunks */
					int (*sizefcn) (ptr) /* BATatoms[].atomLen function */
    );

gdk_export var_t HEAP_malloc(Heap *heap, size_t nbytes);
gdk_export void HEAP_free(Heap *heap, var_t block);
gdk_export size_t HEAP_private(Heap *h);
gdk_export void HEAP_checkformat(Heap *h);
gdk_export int HEAP_check(Heap *h, HeapRepair *hr);
gdk_export void HEAP_init(Heap *heap, int tpe);
gdk_export int HEAP_mmappable(Heap *heap);

#define HEAP_index(HEAP,INDEX,TYPE)	((TYPE *)((char *) (HEAP)->base + (INDEX)))





#define BATDELETE	-9999

gdk_export BAT *BATnew(int hdtype, int tltype, size_t capacity);
gdk_export BAT *BATclone(BAT *b, size_t capacity);
gdk_export BAT *BATextend(BAT *b, size_t newcap);

/* internal */
gdk_export BAT *BATnewstorage(int ht, int tt, size_t cap);
gdk_export BAT *BATcreatedesc(int ht, int tt, int heapnames);
gdk_export int BATfree(BAT *b);
gdk_export void BATdestroy(BAT *b);
gdk_export chr 	ATOMelmshift(int sz);
gdk_export void BATsetdims(BAT *b);



#define hfastins_nocheck(b, p, v, s) {				\
		(b)->H->heap.free += s;				\
		ATOMput((b)->htype, (b)->H->vheap, Hloc(b,p), v);	\
	}
#define tfastins_nocheck(b, p, v, s) {				\
		(b)->T->heap.free += s;				\
		ATOMput((b)->ttype, (b)->T->vheap, Tloc(b,p), v);	\
	}

#define bunfastins_nocheck(b, p, h, t, hs, ts) {		\
		(b)->H->heap.free += hs;				\
		ATOMput((b)->htype, (b)->H->vheap, Hloc(b,p), h);	\
		(b)->T->heap.free += ts;				\
		ATOMput((b)->ttype, (b)->T->vheap, Tloc(b,p), t);	\
		(b)->batCount++;				\
	}

#define bunfastins(b, h, t) {					\
		REGISTER BUN _p = BUNlast(b);			\
		REGISTER size_t _hsize = Hsize(b);		\
		REGISTER size_t _tsize = Tsize(b);		\
		if (BUNlast(b) + 1 > BATcapacity(b)){		\
			if (BATextend(b, BATgrows(b)) == NULL)  \
				goto bunins_failed; 		\
		}						\
		hfastins_nocheck(b, _p, h, _hsize);		\
		tfastins_nocheck(b, _p, t, _tsize);		\
		(b)->batCount++;				\
	}

#define bunfastins_check(b, p, h, t) {                                  \
		bunfastins(b, h, t);                                    \
	}

#define bunfastins_nocheck_inc(b, p, h, t) {                            \
		hfastins_nocheck(b, p, h, Hsize(b));			\
		tfastins_nocheck(b, p, t, Tsize(b));			\
		p++;							\
		(b)->batCount++;					\
	}

gdk_export BAT *BUNfastins(BAT *b, ptr left, ptr right);
gdk_export BAT *BUNins(BAT *b, ptr left, ptr right, bit force);
gdk_export BAT *BUNappend(BAT *b, ptr right, bit force);
gdk_export BAT *BATins(BAT *b, BAT *c, bit force);
gdk_export BAT *BATappend(BAT *b, BAT *c, bit force);
gdk_export BAT *BUNdel(BAT *b, ptr left, ptr right, bit force);
gdk_export BAT *BUNdelHead(BAT *b, ptr left, bit force);
gdk_export BUN BUNdelete(BAT *b, BUN p, bit force);
gdk_export BAT *BATdel(BAT *b, BAT *c, bit force);
gdk_export BAT *BATdelHead(BAT *b, BAT *c, bit force);

gdk_export BAT *BUNreplace(BAT *b, ptr left, ptr right, bit force);
gdk_export BAT *BUNinplace(BAT *b, BUN p, ptr left, ptr right, bit force);
gdk_export BAT *BATreplace(BAT *b, BAT *n, bit force);

gdk_export BUN BUNlocate(BAT *b, ptr left, ptr right);
gdk_export BUN BUNfnd(BAT *b, ptr left);

#define BUNfndVOID(p,bi,v) {                                            \
    BUN result = BUNfirst((bi).b) + (*(oid*)(v) - (bi).b->hseqbase);	\
    ssize_t check = 							\
	(((*(oid*)(v) == oid_nil) ^ ((bi).b->hseqbase == oid_nil)) |    \
        (*(oid*) v < (bi).b->hseqbase) |                           	\
        (*(oid*) v >= (bi).b->hseqbase + (bi).b->batCount));       	\
    p = (check)?BUN_NONE:result; /* and with 0xFF...FF or 0x00..00 */   \
}

#if SIZEOF_OID == SIZEOF_INT
#define BUNfndOID(p,bi,v)			\
	if (BAThdense(bi.b)) {			\
		BUNfndVOID(p,bi,v);		\
	} else {				\
		HASHfnd_int(p,bi,(int*)v);	\
	}
#else
#define BUNfndOID(p,bi,v)			\
	if (BAThdense(bi.b)) {			\
		BUNfndVOID(p,bi,v);		\
	} else {				\
		HASHfnd_lng(p,bi,(lng*)v);	\
	}
#endif
#define BUNfndSTD(p,bi,v) ((p) = BUNfnd(bi.b,v))

#define BAThtype(b)	((b)->htype == TYPE_void && (b)->hseqbase == oid_nil ?\
				TYPE_void : ATOMtype((b)->htype))
#define BATttype(b)	((b)->ttype == TYPE_void && (b)->tseqbase == oid_nil ?\
				TYPE_void : ATOMtype((b)->ttype))
#define BAThstore(b)	(BAThdense(b) ? TYPE_void : (b)->htype)
#define BATtstore(b)	(BATtdense(b) ? TYPE_void : (b)->ttype)
#define Hbase(b)	((b)->H->vheap->base)
#define Tbase(b)	((b)->T->vheap->base)

#define Hsize(b)	((b)->H->width)
#define Tsize(b)	((b)->T->width)

/* new semantics ! */
#define headsize(b,p)   (b->H->type?((size_t)(p))<<b->H->shift:0)
#define tailsize(b,p)   (b->T->type?((size_t)(p))<<b->T->shift:0)

#define Hloc(b,p)	((b)->H->heap.base+((p)<<(b)->H->shift))
#define Tloc(b,p)	((b)->T->heap.base+((p)<<(b)->T->shift))

#define BUNhloc(bi,p)	Hloc((bi).b,p)
#define BUNtloc(bi,p)	Tloc((bi).b,p)
#define BUNhpos(bi,p)	(Hpos(&(bi),p))
#define BUNtpos(bi,p)	(Tpos(&(bi),p))
#define BUNhvar(bi,p)	((bi).b->htype?(Hbase((bi).b)+*((var_t*)BUNhloc(bi,p))):BUNhpos(bi,p))
#define BUNtvar(bi,p)	((bi).b->ttype?(Tbase((bi).b)+*((var_t*)BUNtloc(bi,p))):BUNtpos(bi,p))
#define BUNhead(bi,p)	((bi).b->hvarsized?BUNhvar(bi,p):BUNhloc(bi,p))
#define BUNtail(bi,p)	((bi).b->tvarsized?BUNtvar(bi,p):BUNtloc(bi,p))

static INLINE BATiter bat_iterator(BAT *b) 
{
	BATiter bi;

	bi.b = b;
	bi.hvid = bi.tvid = 0;
	return bi;
}

#define BUNfirst(b)	((b)->batFirst)
#define BUNlast(b)	((b)->batFirst + (b)->batCount)



gdk_export str BATrename(BAT *b, str nme);
gdk_export size_t BATcount(BAT *b);
gdk_export size_t BATcount_no_nil(BAT *b);
gdk_export void BATsetcapacity(BAT *b, size_t cnt);
gdk_export void BATsetcount(BAT *b, size_t cnt);
gdk_export size_t BATbuncount(BAT *b);
gdk_export size_t BATguess(BAT *b);
gdk_export size_t BATgrows(BAT *b);
gdk_export BAT *BATkey(BAT *b, int onoff);
gdk_export BAT *BATset(BAT *b, int onoff);
gdk_export BAT *BATmode(BAT *b, int onoff);
gdk_export BAT *BATroles(BAT *b, str hnme, str tnme);
gdk_export BAT *BATcol_name(BAT *b, str tnme);
gdk_export int BATname(BAT *b, str nme);
gdk_export BAT *BATseqbase(BAT *b, oid o);
gdk_export BAT *BATsetaccess(BAT *b, int mode);
gdk_export int BATgetaccess(BAT *b);
gdk_export int BATcheckmodes(BAT *b, int persistent, int unloadable);


#define BATdirty(b)	((b)->batCopiedtodisk == 0 || (b)->batDirty || \
			(b)->batDirtydesc || \
			(b)->H->heap.dirty || (b)->T->heap.dirty || \
			((b)->H->vheap?(b)->H->vheap->dirty:0) || \
			((b)->T->vheap?(b)->T->vheap->dirty:0))

#define PERSISTENT	4096
#define SESSION		2048
#define TRANSIENT	1024

#define BAT_WRITE		0	/* all kinds of access allowed */
#define BAT_READ		1	/* only read-access allowed */
#define BAT_APPEND		2	/* only reads and appends allowed */

#define BATversion(b)	(b)->GDKversion
#define BATcapacity(b)	(b)->batCapacity


gdk_export BAT *BATclear(BAT *b);
gdk_export BAT *BATcopy(BAT *b, int ht, int tt, int writeable);
gdk_export BAT *BATmark(BAT *b, oid base);
gdk_export BAT *BATmark_grp(BAT *b, BAT *g, oid *base);
gdk_export BAT *BATnumber(BAT *b);
gdk_export BAT *BATgroup(BAT *b, int start, int incr, int grpsize);



/* Heap storage modes */
#define STORE_MEM	0	/* load into GDKmalloced memory */
#define STORE_MMAP	1	/* mmap() into virtual memory */
#define STORE_PRIV	2	/* BAT copy of copy-on-write mmap */
#define STORE_UNSET	((char)-1)	/* no mode specified */

gdk_export size_t GDK_mem_pagebits;	/* page size for non-linear mmaps */

#define REMAP_PAGE_BITS    GDK_mem_pagebits
#define REMAP_PAGE_SIZE    (1 << REMAP_PAGE_BITS)
#define REMAP_PAGE_MASK    (REMAP_PAGE_SIZE - 1)
#define REMAP_PAGE_MAXBITS (REMAP_PAGE_BITS+3)
#define REMAP_PAGE_MAXSIZE (1<<REMAP_PAGE_MAXBITS)	/* max page bytesize of unary BUN heap (8-byte atom) */

/* Buffer management advise for heaps */
#define BUF_NORMAL	0	/* No further special treatment */
#define BUF_RANDOM	1	/* Expect random page references */
#define BUF_SEQUENTIAL	2	/* Expect sequential page references */
#define BUF_WILLNEED	3	/* Will need these pages */
#define BUF_DONTNEED	4	/* Don't need these pages */

gdk_export BAT *BATload(str nme);
gdk_export BAT *BATload_intern(bat bid);
gdk_export BAT *BATsave(BAT *b);
gdk_export int BATmmap(BAT *b, int bn, int hh, int th);
gdk_export int BATmmap_pin(BAT *b);
gdk_export int BATmmap_unpin(BAT *b);
gdk_export int BATmadvise(BAT *b, int bn, int hh, int th);
gdk_export int BATdelete(BAT *b);
gdk_export size_t BATvmsize(BAT *b, int dirty);
gdk_export size_t BATmemsize(BAT *b, int dirty);

gdk_export int GDKfdlocate(str nme, str mode, str ext);
gdk_export FILE *GDKfilelocate(str nme, str mode, str ext);
gdk_export void GDKfilepath(str path, str nme, str mode, str ext);
gdk_export int GDKcreatedir(str nme);
gdk_export int GDKremovedir(str nme);

gdk_export char *GDKload(str nme, str ext, size_t size, size_t chunk, int mode);
gdk_export int GDKsave(str nme, str ext, void *buf, size_t size, int mode);
gdk_export int GDKunlink(str dir, str nme, str extension);
gdk_export int GDKmove(str dir1, str nme1, str ext1, str dir2, str nme2, str ext2);


gdk_export int BATprint(BAT *b);
gdk_export int BATprintf(stream *f, BAT *b);
gdk_export int BATmultiprintf(stream *f, int argc, BAT *argv[], int printoid, int order, int printorderby);



#define GDK_SORTED_REV	128	/* reversely sorted */
#define GDK_SORTED	65	/* 65 = (32 bits radix_clustered)<<1 + 1 */
#define REVERT_SORTED(o) \
	o==GDK_SORTED?GDK_SORTED_REV:(o==GDK_SORTED_REV?GDK_SORTED:0)

gdk_export BAT *BATsort(BAT *b);
gdk_export BAT *BATsort_rev(BAT *b);
gdk_export BAT *BATorder(BAT *b);
gdk_export BAT *BATorder_rev(BAT *b);
gdk_export BAT *BATrevert(BAT *b);
gdk_export int BATordered(BAT *b);
gdk_export int BATordered_rev(BAT *b);
gdk_export BAT *BATssort(BAT *b);
gdk_export BAT *BATssort_rev(BAT *b);
gdk_export BAT *BATsorder(BAT *b);
gdk_export BAT *BATsorder_rev(BAT *b);

gdk_export void GDKqsort(void *h, void *t, void *base, size_t n, int hs, int ts, int tpe); 
gdk_export void GDKqsort_rev(void *h, void *t, void *base, size_t n, int hs, int ts, int tpe); 

gdk_export int GDKssort(void *h, void *t, void *base, size_t n, int hs, int ts, int tpe); 
gdk_export int GDKssort_rev(void *h, void *t, void *base, size_t n, int hs, int ts, int tpe); 

#define BAThordered(b)	(((b)->htype == TYPE_void)?GDK_SORTED:(b)->hsorted)
#define BATtordered(b)	(((b)->ttype == TYPE_void)?GDK_SORTED:(b)->tsorted)
#define BAThdense(b)	(BAThvoid(b) && (b)->hseqbase != oid_nil)
#define BATtdense(b)	(BATtvoid(b) && (b)->tseqbase != oid_nil)
#define BAThvoid(b)	(((b)->hdense&(b)->hsorted&1) || (b)->htype==TYPE_void)
#define BATtvoid(b)	(((b)->tdense&(b)->tsorted&1) || (b)->ttype==TYPE_void)
#define BAThkey(b)	(b->hkey != FALSE || BAThdense(b))
#define BATtkey(b)	(b->tkey != FALSE || BATtdense(b))

/* set some properties that are trivial to deduce */
#define BATsettrivprop(b)	do { \
					int trivial = (b)->batCount <= 1; \
					if (!(b)->hdense && (b)->htype == TYPE_void && (b)->hseqbase != oid_nil) { \
						(b)->hdense = 1; \
						(b)->batDirtydesc = TRUE; \
					} \
					if (!(b)->tdense && (b)->ttype == TYPE_void && (b)->tseqbase != oid_nil) { \
						(b)->tdense = 1; \
						(b)->batDirtydesc = TRUE; \
					} \
					if (trivial) { \
						oid sqbs; \
						BATiter bi = bat_iterator(b); \
						if (!(b)->hdense && \
						    (b)->htype == TYPE_oid && \
						    (sqbs = (b)->batCount == 0 ? 0 : * (oid *) BUNhead(bi, BUNfirst(b))) != oid_nil) { \
							(b)->batDirtydesc = TRUE; \
							(b)->hdense = 1; \
							BATseqbase((b), sqbs); \
						} \
						if (!(b)->tdense && \
						    (b)->ttype == TYPE_oid && \
						    (sqbs = (b)->batCount == 0 ? 0 : * (oid *) BUNtail(bi, BUNfirst(b))) != oid_nil) { \
							(b)->batDirtydesc = TRUE; \
							(b)->tdense = 1; \
							BATseqbase(BATmirror(b), sqbs); \
						} \
					} \
					if ((trivial && (b)->htype != TYPE_void) || (b)->hdense) { \
						if ((b)->hsorted != GDK_SORTED) \
							(b)->batDirtydesc = TRUE; \
						(b)->hsorted = GDK_SORTED; \
						if (!(b)->hkey) \
							BATkey((b), TRUE); \
					} \
					if ((trivial && (b)->ttype != TYPE_void) || (b)->tdense) { \
						if ((b)->tsorted != GDK_SORTED) \
							(b)->batDirtydesc = TRUE; \
						(b)->tsorted = GDK_SORTED; \
						if (!(b)->tkey) \
							BATkey(BATmirror(b), TRUE); \
					} \
				} while (0)


typedef struct {
	BAT *b[2];		/* if loaded: BAT* handle + reverse */
	str nme[2];		/* logical name + reverse */
	str bak[2];		/* logical name + reverse backups */
	bat next[2];		/* next BBP slot in link list */
	BATstore *cache;	/* cached header info */
	str path;		/* dir + basename for storage */
	int refs;		/* in-memory references on which the loaded status of a BAT relies */
	int lrefs;		/* logical references on which the existence of a BAT relies */
	int lastused;		/* BBP LRU stamp */
	volatile int status;	/* status mask used for spin locking */
	/* MT_Id pid;           non-zero thread-id if this BAT is private */
} BBPrec;

gdk_export bat BBPmaxsize;
gdk_export bat BBPlimit;
gdk_export int BBP_dirty;	/* BBP table dirty? */
gdk_export BBPrec *BBP;

/* fast defines without checks; internal use only  */
#define BBP_cache(i)    BBP[ABS(i)].b[(i)<0]
#define BBP_logical(i)  BBP[ABS(i)].nme[(i)<0]
#define BBP_next(i)     BBP[ABS(i)].next[(i)<0]
#define BBP_physical(i) BBP[ABS(i)].path
#define BBP_desc(i)	BBP[ABS(i)].cache
#define BBP_refs(i)     BBP[ABS(i)].refs
#define BBP_lrefs(i)    BBP[ABS(i)].lrefs
#define BBP_lastused(i) BBP[ABS(i)].lastused
#define BBP_status(i)   BBP[ABS(i)].status
#define BBP_pid(i)      BBP[ABS(i)].pid

/* macros that nicely check parameters */
#define BBPdirty(x)	 (BBP_dirty=(x))
#define BBPcacheid(b)	 ((b)->batCacheid)
#define BBPstatus(i)	 (BBPcheck((i),"BBPstatus")?BBP_status(i):-1)
#define BBPcurstamp()	 BBP_curstamp
#define BBPrefs(i)	 (BBPcheck((i),"BBPrefs")?BBP_refs(i):-1)
#define BBPcache(i)	 (BBPcheck((i),"BBPcache")?BBP_cache(i):(BAT*) NULL)
#define BBPname(i)	 (BBPcheck((i),"BBPname")?((i) > 0 || BBP_logical(i))?BBP_logical(i):BBP_logical(-(i)):"")
#define BBPvalid(i)      (BBP_logical(i) != NULL && *BBP_logical(i) != '.')
#define BATgetId(b)	 BBPname((b)->batCacheid)
#define BBPfix(i)	 BBPincref(i, FALSE)
#define BBPunfix(i)	 BBPdecref(i, FALSE)

#define BBPRENAME_ALREADY	-1
#define BBPRENAME_ILLEGAL	-2
#define BBPRENAME_LONG		-3

gdk_export void BBPlock(str s);

gdk_export void BBPhot(bat b);
gdk_export void BBPcold(bat b);
gdk_export void BBPtrim(size_t memdelta, size_t vmdelta);
gdk_export void BBPunlock(str s);

gdk_export str BBPlogical(bat b, str buf);
gdk_export str BBPphysical(bat b, str buf);
gdk_export int BBP_curstamp;
gdk_export BATstore *BBPgetdesc(bat i);
gdk_export BAT *BBPquickdesc(bat b, int delaccess);



typedef struct {
	/* simple attributes */
	char name[IDLENGTH];
	int storage;		/* stored as another type? */
	short linear;		/* atom can be ordered linearly */
	short size;		/* fixed size of atom */
	short align;		/* alignment condition for values */
	short deleting;		/* set if unloading */
	int varsized;		/* variable-size or fixed-sized */

	/* automatically generated fields */
	ptr atomNull;		/* global nil value */

	/* generic (fixed + varsized atom) ADT functions */
	int (*atomFromStr) (str s, int *len, ptr *dst);
	int (*atomToStr) (str *s, int *len, ptr src);
	void *(*atomRead) (ptr a, stream *s, size_t cnt);
	int (*atomWrite) (ptr a, stream *s, size_t cnt);
	int (*atomCmp) (ptr v1, ptr v2);
	hash_t (*atomHash) (ptr v);
	/* optional functions */
	void (*atomConvert) (ptr v, int direction);
	int (*atomFix) (ptr atom);
	int (*atomUnfix) (ptr atom);

	/* varsized atom-only ADT functions */
	var_t (*atomPut) (Heap *, var_t *off, ptr src);
	void (*atomDel) (Heap *, var_t *atom);
	int (*atomLen) (ptr atom);
	void (*atomHeap) (Heap *, size_t);
	/* optional functions */
	void (*atomHeapConvert) (Heap *, int direction);
	int (*atomHeapCheck) (Heap *, HeapRepair *);
} atomDesc;

gdk_export atomDesc BATatoms[];
gdk_export int GDKatomcnt;

gdk_export void ATOMdelete(int id);
gdk_export void ATOMproperty(char *nme, char *property, GDKfcn fcn);
gdk_export int ATOMindex(char *nme);
gdk_export int ATOMisdescendant(int id, int parentid);
gdk_export int ATOMdump(void);

gdk_export int ATOMheap(int id, Heap *hp, size_t cap);
gdk_export str ATOMname(int id);
gdk_export int ATOMlen(int id, ptr v);
gdk_export ptr ATOMnil(int id);
gdk_export int ATOMcmp(int id, ptr v_1, ptr v_2);
gdk_export int ATOMprint(int id, ptr val, stream *fd);
gdk_export int ATOMformat(int id, ptr val, char **buf);

gdk_export ptr ATOMdup(int id, ptr val);



gdk_export int OIDinit(void);
gdk_export oid OIDrand(void);
gdk_export oid OIDbase(oid base);
gdk_export oid OIDseed(oid seed);
gdk_export oid OIDnew(size_t inc);
gdk_export oid OIDread(str buf);
gdk_export int OIDwrite(stream *fp);
gdk_export int OIDdirty(void);



gdk_export BAT *BAThash(BAT *b, size_t masksize);
gdk_export BAT *BAThashsplit(BAT *b, size_t n, int unary);
gdk_export BAT *BATrangesplit(BAT *b, size_t n, int unary);
gdk_export BAT *BAThashjoin(BAT *l, BAT *r, size_t estimate);

/* low level functions */
gdk_export BUN SORTfnd(BAT *b, ptr v);
gdk_export BUN SORTfndfirst(BAT *b, ptr v);
gdk_export BUN SORTfndlast(BAT *b, ptr v);

#define BATprepareHash(X) (((X)->H->hash == NULL) && !BAThash(X, 0))


#define GDK_HISTO_MAX_BIT             ((int) (sizeof(size_t)<<3))

/* we prefer to use vm_alloc routines on size > GDKmmap */
gdk_export void GDKvminc(size_t len);
gdk_export void *GDKmmap(char *path, int mode, off_t off, size_t len);
gdk_export int GDKmunmap(void *addr, size_t len);

gdk_export ptr GDK_mem_start;		/* sbrk(0) at start of the program */
gdk_export size_t GDK_mem_bigsize;	/* size after which we use VM rather than heap */
gdk_export size_t GDK_mem_maxsize;	/* max allowed size of committed memory */
gdk_export size_t GDK_vm_minsize;	/* smallest size allowed for a vm block */
gdk_export size_t GDK_vm_maxsize;	/* max allowed size of reserverd vm */
gdk_export int    GDK_vm_trim;		/* run vmtrim thread */

gdk_export size_t GDKmem_inuse(void);	/* RAM/swapmem that MonetDB is really using now */
gdk_export size_t GDKmem_cursize(void);	/* RAM/swapmem that MonetDB has claimed from OS */
gdk_export size_t GDKvm_cursize(void);	/* current MonetDB VM address space usage */
gdk_export size_t GDKvm_heapsize(void);
gdk_export size_t GDKmem_heapsize(void);
gdk_export size_t GDKmem_heapinuse(void);

gdk_export void *GDKmalloc(size_t size);
gdk_export void *GDKzalloc(size_t size);
gdk_export void *GDKmallocmax(size_t size, size_t * maxsize, int emergency);
gdk_export void *GDKrealloc(void *pold, size_t size);
gdk_export void *GDKreallocmax(void *pold, size_t size, size_t * maxsize, int emergency);
gdk_export void GDKfree(void *blk);
gdk_export str GDKstrdup(const char *s);

gdk_export void *GDKvmalloc(size_t size, size_t * maxsize, int emergency);
gdk_export void *GDKvmrealloc(void *pold, size_t oldsize, size_t newsize, size_t oldmax, size_t * maxsize, int emergency);
gdk_export void GDKvmfree(void *blk, size_t size, size_t maxsize);



#define GDKMAXERRLEN	10240
#define GDKWARNING	"!WARNING: "
#define GDKERROR	"!ERROR: "
#define GDKMESSAGE	"!OS: "
#define GDKFATAL	"!FATAL: "

/* Data Distilleries uses ICU for internationalization of some MonetDB error messages */
#ifdef MONET_ICU_I18N
gdk_export void GDKvsnprintf(char *buf, int maxlen, const char *format, va_list ap);
#else
#define GDKvsnprintf(buf, maxlen, format, ap) vsnprintf(buf, maxlen, format, ap)
#endif

gdk_export void GDKaddbuf(const char *msg);
gdk_export int GDKwarning(const char *format, ...);
gdk_export int GDKerror(const char *format, ...);
gdk_export int GDKsyserror(const char *format, ...);
gdk_export int GDKfatal(const char *format, ...);
gdk_export int GDKerrorCount(void);



#include "gdk_delta.h"
#include "gdk_search.h"
#include "gdk_atoms.h"
#include "gdk_bbp.h"
#include "gdk_desc.h"
#include "gdk_utils.h"
#include "gdk_bat.h"

#ifdef NATIVE_WIN32
#ifdef _MSC_VER
#define fileno _fileno
#endif
#define fdopen _fdopen
#define putenv _putenv
#define strdup _strdup
#endif

#ifndef S_IRUSR
/* if one doesn't exist, presumably they all don't exist */
#define S_IRUSR 0000400		/* read permission, owner */
#define S_IWUSR 0000200		/* write permission, owner */
#define S_IRGRP 0000040		/* read permission, group */
#define S_IWGRP 0000020		/* write permission, grougroup */
#define S_IROTH 0000004		/* read permission, other */
#define S_IWOTH 0000002		/* write permission, other */
#endif

#define MONETDB_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)

#define BBP_BATMASK     511
#define BBP_THREADMASK   63

typedef struct {
	MT_Lock swap;
	MT_Lock hash;
} batlock_t;

typedef struct {
	MT_Lock alloc;
	MT_Lock trim;
	bat free;
} bbplock_t;

#define GDKswapLock(x)  GDKbatLock[(x&BBP_BATMASK)].swap
#define GDKhashLock(x)  GDKbatLock[(x&BBP_BATMASK)].hash
#define GDKtrimLock(y)  GDKbbpLock[(y&BBP_THREADMASK)].trim
#define GDKcacheLock(y) GDKbbpLock[(y&BBP_THREADMASK)].alloc
#define BBP_free(y)     GDKbbpLock[(y&BBP_THREADMASK)].free

gdk_export batlock_t GDKbatLock[BBP_BATMASK + 1];
gdk_export bbplock_t GDKbbpLock[BBP_THREADMASK + 1];
gdk_export MT_Lock GDKnameLock;
gdk_export MT_Lock GDKthreadLock;
gdk_export MT_Lock GDKunloadLock;
gdk_export MT_Cond GDKunloadCond;
gdk_export MT_Lock GDKtmLock;

#define CHECKMASK         2
#define CHECKDEBUG      if (GDKdebug & CHECKMASK)
#define MEMMASK           4
#define MEMDEBUG        if (GDKdebug & MEMMASK)
#define PROPMASK          8
#define PROPDEBUG       if (GDKdebug & PROPMASK)
#define IOMASK            16
#define IODEBUG         if (GDKdebug & IOMASK)
#define BATMASK          32
#define BATDEBUG        if (GDKdebug & BATMASK)
#define PARSEMASK         64
#define PARSEDEBUG      if (GDKdebug & PARSEMASK)
#define PARMASK           128
#define PARDEBUG        if (GDKdebug & PARMASK)
#define TRGMASK           256
#define TRGDEBUG        if (GDKdebug & TRGMASK)
#define TMMASK            512
#define TMDEBUG         if (GDKdebug & TMMASK)
#define TEMMASK           1024
#define TEMDEBUG        if (GDKdebug & TEMMASK)
#define DLMASK	           2048
#define DLDEBUG	        if (GDKdebug & DLMASK)
#define PERFMASK	       4096
#define PERFDEBUG	    if (GDKdebug & PERFMASK)
#define DELTAMASK	       8192
#define DELTADEBUG	    if (GDKdebug & DELTAMASK)
#define LOADMASK	      16384
#define LOADDEBUG	    if (GDKdebug & LOADMASK)
#define YACCMASK	       32768
#define YACCDEBUG	    if (GDKdebug & YACCMASK)
/*
#define ?tcpip?		if (GDKdebug&65536)
#define ?monet_multiplex?	if (GDKdebug&131072)
#define ?ddbench?	if (GDKdebug&262144)
#define ?ddbench?	if (GDKdebug&524288)
#define ?ddbench?	if (GDKdebug&1048576)
*/
#define ALGOMASK	    2097152
#define ALGODEBUG	    if (GDKdebug & ALGOMASK)
#define ESTIMASK		4194304
#define ESTIDEBUG		if (GDKdebug & ESTIMASK)
#define XPROPMASK      	8388608
#define XPROPDEBUG      if (GDKdebug & XPROPMASK)

#define JOINPROPMASK     16777216
#define JOINPROPCHK     if (!(GDKdebug & JOINPROPMASK))
#define DEADBEEFMASK     33554432
#define DEADBEEFCHK     if (!(GDKdebug & DEADBEEFMASK))

gdk_export int GDKnrofthreads;
gdk_export int GDKembedded;
gdk_export int GDKprotected;
gdk_export void GDKprotect(void);





#define gdk_set_lock(X,Y)    if(GDKprotected ) MT_set_lock(X,Y)
#define gdk_unset_lock(X,Y)  if(GDKprotected ) MT_unset_lock(X,Y)
#define gdk_up_sema(X,Y)     if(GDKprotected ) MT_up_sema(X,Y)
#define gdk_down_sema(X,Y)   if(GDKprotected ) MT_down_sema(X,Y)
#define gdk_signal_cond(X,Y) if(GDKprotected ) MT_signal_cond(X,Y)
#define gdk_wait_cond(X,Y,Z) if(GDKprotected ) MT_wait_cond(X,Y,Z)




#define THREADS	64
#define THREADDATA	16

typedef struct threadStruct {
	int tid;		/* logical ID by MonetDB; val == index into this array + 1 (0 is invalid) */
	MT_Id pid;		/* physical thread id (pointer-sized) from the OS thread library */
	str name;
	ptr data[THREADDATA];
	size_t sp;
	void (*cleanup) (ptr thr);
} ThreadRec, *Thread;


gdk_export ThreadRec GDKthreads[THREADS];

gdk_export int THRgettid(void);
gdk_export Thread THRget(int tid);
gdk_export Thread THRnew(MT_Id pid, str name);
gdk_export void THRdel(Thread t);
gdk_export int THRinit(void);
gdk_export int THRexit(str nme);
gdk_export int THRcnt(void);
gdk_export void THRsetdata(int, ptr);
gdk_export void *THRgetdata(int);
gdk_export int THRhighwater(void);
gdk_export int THRprintf(stream *s, const char *format, ...);

gdk_export void *THRdata[16];

#define GDKstdout	((stream*)THRdata[0])
#define GDKstdin	((stream*)THRdata[1])

#define GDKout	        ((stream*)THRgetdata(0))
#define GDKin	        ((stream*)THRgetdata(1))
#define GDKerrbuf	((char*)THRgetdata(2))
#define GDKsetbuf(x)	THRsetdata(2,(ptr)(x))
#define GDKerr	        GDKout

#define THRget_errbuf(t)	((char*)t->data[2])
#define THRset_errbuf(t,b)	(t->data[2] = b)

#ifndef GDK_NOLINK
#ifdef __cplusplus
extern "C" {
#endif

static INLINE bat
BBPcheck(REGISTER bat x, REGISTER const char *y)
{
	if (x && x != bat_nil) {
		REGISTER bat z = ABS(x);

		if (z >= BBPsize || BBP[z].nme[0] == NULL) {
			GDKwarning("%s: range error %d\n", y, x);
		} else {
			return z;
		}
	}
	return 0;
}

static INLINE BAT *
BATdescriptor(REGISTER bat i)
{
	REGISTER BAT *b = NULL;

	if (BBPcheck(i, "BATdescriptor")) {
		BBPfix(i);
		b = BBP_cache(i);
		if (b == NULL)
			b = BBPdescriptor(i);
	}
	return b;
}

static INLINE char *
Hpos(BATiter *bi, BUN p)
{
	bi->hvid = bi->b->hseqbase;
	if (bi->hvid != oid_nil) 
		bi->hvid += p - BUNfirst(bi->b); 
	return (char*)&bi->hvid;
}

static INLINE char *
Tpos(BATiter *bi, BUN p)
{
	bi->tvid = bi->b->tseqbase;
	if (bi->tvid != oid_nil) 
		bi->tvid += p - BUNfirst(bi->b); 
	return (char*)&bi->tvid;
}

static INLINE BAT *
BATmirror(REGISTER BAT *b)
{
	if (b == NULL)
		return NULL;
	return BBP_cache(-b->batCacheid);
}

#ifdef __cplusplus
}
#endif
#endif



gdk_export int TMcommit(void);
gdk_export int TMabort(void);
gdk_export int TMsubcommit(BAT *bl);



gdk_export BAT *BATcommit(BAT *b);
gdk_export BAT *BATfakeCommit(BAT *b);
gdk_export BAT *BATundo(BAT *b);
gdk_export BAT *BATalpha(BAT *b);
gdk_export BAT *BATdelta(BAT *b);
gdk_export BAT *BATprev(BAT *b);





gdk_export int ALIGNsynced(BAT *b1, BAT *b2);
gdk_export int ALIGNrelated(BAT *b1, BAT *b2);

gdk_export BAT *BATpropcheck(BAT *b, int mode);

#define BATPROPS_QUICK  0	/* only derive easy (non-resource consuming) properties */
#define BATPROPS_ALL    1	/* derive all possible properties; no matter what cost (key=hash) */
#define BATPROPS_CHECK  3	/* BATPROPS_ALL, but start from scratch and report illegally set properties */

gdk_export BAT *VIEWcreate(BAT *h, BAT *t);
gdk_export BAT *VIEWcreate_(BAT *h, BAT *t, int stable);
gdk_export BAT *VIEWhead(BAT *b);
gdk_export BAT *VIEWhead_(BAT *b, int mode);
gdk_export BAT *VIEWcombine(BAT *b);
gdk_export BAT *VIEWreset(BAT *b);
gdk_export BAT *BATmaterialize(BAT *b);
gdk_export BAT *BATmaterializeh(BAT *b);
gdk_export BAT *BATmaterializet(BAT *b);
gdk_export void VIEWdestroy(BAT *b);
gdk_export void VIEWbounds(BAT *view, size_t l, size_t h);
gdk_export void VIEWunlink(BAT *b);

/* low level functions */
gdk_export int ALIGNundo(BAT *b);
gdk_export int ALIGNcommit(BAT *b);
gdk_export int ALIGNsetH(BAT *b1, BAT *b2);

#define ALIGNset(x,y)	{ALIGNsetH(x,y);ALIGNsetT(x,y);}
#define ALIGNsetT(x,y)	ALIGNsetH(BATmirror(x),BATmirror(y))
#define ALIGNins(x,y,f)	{if (!(f)) VIEWchk(x,y,BAT_READ);(x)->halign=(x)->talign=0; }
#define ALIGNdel(x,y,f)	{if (!(f)) VIEWchk(x,y,BAT_READ|BAT_APPEND);(x)->halign=(x)->talign=0; }
#define ALIGNinp(x,y,f) {if (!(f)) VIEWchk(x,y,BAT_READ|BAT_APPEND);(x)->talign=0; }
#define ALIGNapp(x,y,f) {if (!(f)) VIEWchk(x,y,BAT_READ);(x)->talign=0; }

#define BAThrestricted(b) (VIEWhparent(b) ? BBP_cache(VIEWhparent(b))->batRestricted : (b)->batRestricted)
#define BATtrestricted(b) (VIEWtparent(b) ? BBP_cache(VIEWtparent(b))->batRestricted : (b)->batRestricted)

/* The batRestricted field indicates whether a BAT is readonly.
 * we have modes: BAT_WRITE  = all permitted
 *                BAT_APPEND = append-only
 *                BAT_READ   = read-only
 * VIEW bats are always mapped read-only.
 */
#define	VIEWchk(x,y,z)							  	\
	if ((((x)->batRestricted & (z)) != 0) | ((x)->batSharecnt > 0)) {	 \
		GDKerror("%s: access denied to %s, aborting.\n", y, BATgetId(x));\
		return 0;							 \
	}

/* the parentid in a VIEW is correct for the normal view. We must correct
 * for the reversed view. A special case are the VIEWcombine bats, these
 * always refer to the same parent column (i.e. no correction needed)
 */
#define isVIEW(x) (!(x)->P->lview && ((x)->H->parentid || (x)->T->parentid))
#define isVIEWCOMBINE(x) ((x)->H == (x)->T)
#define VIEWhparent(x)	((x)->P->lview?0:(x)->H->parentid)
#define VIEWtparent(x)	((x)->P->lview?0:(x)->T->parentid)

/* VIEWparentcol(b) tells whether the head column was inherited from the parent
 * "as is". We must check whether the type was not overridden in the view.
 */
#define VIEWparentcol(b) ((VIEWhparent(b) && (b)->htype			\
			   && (b)->htype == BBP_cache(VIEWhparent(b))->htype) \
			  ?VIEWhparent(b):0)


#define BATloop(r, p, q) \
	for(q = BUNlast(r), p = BUNfirst(r);p < q; p++)



#define BATloopDEL(r, p, q)						\
	for(p = BUNfirst(r), q = BUNlast(r); p < q;	\
	    q = MIN(q,BUNlast(r)), p++)



#define DELloop(b, p, q)						\
	for (q = (b)->batFirst, p = (b)->batDeleted; p < q; p++)



#define GDK_STREQ(l,r) (*(char*) (l) == *(char*) (r) && !strcmp(l,r))

#define HASHloop(bi, h, hb, v)						\
	for (hb = h->hash[HASHprobe(h, v)]; hb != HASH_MAX; hb = h->link[hb])	\
		if (ATOMcmp(h->type, v, BUNhead(bi, hb)) == 0)
#define HASHloop_str(bi, h, hb, v)					\
	for (hb = h->hash[strHash(v)&h->mask]; hb != HASH_MAX; hb = h->link[hb])	\
		if (GDK_STREQ(v, BUNhvar(bi, hb)))


#define HASHloop_fstr(bi, h, hb, idx, v)					\
	for (hb = h->hash[strHash(v)&h->mask], idx = strLocate((bi.b)->hheap,v); \
	     hb != HASH_MAX; hb = h->link[hb])				\
		     if (*(var_t*) BUNhloc(bi, hb) == idx)


#define HASHlooploc(bi,	h, hb, v)					\
	for (hb = h->hash[HASHprobe(h, v)]; hb != HASH_MAX; hb = h->link[hb])	\
		if (ATOMcmp(h->type, v, BUNhloc(bi, hb)) == 0)
#define HASHloopvar(bi,	h, hb, v)					\
	for (hb = h->hash[HASHprobe(h, v)]; hb != HASH_MAX; hb = h->link[hb])	\
		if (ATOMcmp(h->type, v, BUNhvar(bi, hb)) == 0)



#define HASHloop_bit(bi,h,hb,v)						   \
	for (hb = h->hash[hash_chr(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), bte))




#define HASHloop_chr(bi,h,hb,v)						   \
	for (hb = h->hash[hash_chr(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), chr))




#define HASHloop_bte(bi,h,hb,v)						   \
	for (hb = h->hash[hash_bte(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), bte))




#define HASHloop_sht(bi,h,hb,v)						   \
	for (hb = h->hash[hash_sht(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), sht))




#define HASHloop_int(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))




#if SIZEOF_OID == SIZEOF_INT

#define HASHloop_oid(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))


	/* OIDDEPEND */
#else

#define HASHloop_oid(bi,h,hb,v)						   \
	for (hb = h->hash[hash_lng(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), lng))


	/* OIDDEPEND */
#endif
#if SIZEOF_WRD == SIZEOF_INT

#define HASHloop_wrd(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))



#else

#define HASHloop_wrd(bi,h,hb,v)						   \
	for (hb = h->hash[hash_lng(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), lng))



#endif

#define HASHloop_bat(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))




#if SIZEOF_VOID_P == SIZEOF_INT

#define HASHloop_ptr(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))



#else /* SIZEOF_VOID_P == SIZEOF_LNG */

#define HASHloop_ptr(bi,h,hb,v)						   \
	for (hb = h->hash[hash_lng(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), lng))



#endif

#define HASHloop_flt(bi,h,hb,v)						   \
	for (hb = h->hash[hash_int(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), int))




#define HASHloop_lng(bi,h,hb,v)						   \
	for (hb = h->hash[hash_lng(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), lng))




#define HASHloop_dbl(bi,h,hb,v)						   \
	for (hb = h->hash[hash_lng(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (simple_EQ(v, BUNhloc(bi, hb), lng))




#define HASHloop_any(bi,h,hb,v)						   \
	for (hb = h->hash[hash_any(h,v)]; hb != HASH_MAX; hb = h->link[hb]) \
		if (atom_EQ(v, BUNhead(bi, hb), (bi).b->htype))






#define SORTloop(b,p,q,tl,th)						\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop: BAT not sorted.\n"); \
	else for (p = (ATOMcmp((b)->ttype,tl,ATOMnilptr((b)->ttype))?	\
		     SORTfndfirst(b,tl):BUNfirst(b)),			\
		  q = (ATOMcmp((b)->ttype,th,ATOMnilptr((b)->ttype))?	\
		     SORTfndlast(b,th):BUNlast(b)); p < q; p++)



#define SORTloop_chr(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_chr: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&chr_nil,chr)?BUNfirst(b):SORTfndfirst_chr(b,tl), \
		  q = simple_EQ(th,&chr_nil,chr)?BUNfirst(b):SORTfndlast_chr(b,th); p < q; p++)




#define SORTloop_bte(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_bte: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&bte_nil,bte)?BUNfirst(b):SORTfndfirst_bte(b,tl), \
		  q = simple_EQ(th,&bte_nil,bte)?BUNfirst(b):SORTfndlast_bte(b,th); p < q; p++)




#define SORTloop_sht(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_sht: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&sht_nil,sht)?BUNfirst(b):SORTfndfirst_sht(b,tl), \
		  q = simple_EQ(th,&sht_nil,sht)?BUNfirst(b):SORTfndlast_sht(b,th); p < q; p++)




#define SORTloop_int(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_int: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&int_nil,int)?BUNfirst(b):SORTfndfirst_int(b,tl), \
		  q = simple_EQ(th,&int_nil,int)?BUNfirst(b):SORTfndlast_int(b,th); p < q; p++)




#define SORTloop_flt(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_flt: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&flt_nil,flt)?BUNfirst(b):SORTfndfirst_flt(b,tl), \
		  q = simple_EQ(th,&flt_nil,flt)?BUNfirst(b):SORTfndlast_flt(b,th); p < q; p++)




#define SORTloop_lng(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_lng: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&lng_nil,lng)?BUNfirst(b):SORTfndfirst_lng(b,tl), \
		  q = simple_EQ(th,&lng_nil,lng)?BUNfirst(b):SORTfndlast_lng(b,th); p < q; p++)




#define SORTloop_dbl(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_dbl: BAT not sorted.\n"); \
	else for (p = simple_EQ(tl,&dbl_nil,dbl)?BUNfirst(b):SORTfndfirst_dbl(b,tl), \
		  q = simple_EQ(th,&dbl_nil,dbl)?BUNfirst(b):SORTfndlast_dbl(b,th); p < q; p++)




#define SORTloop_loc(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_loc: BAT not sorted.\n"); \
	else for (p = atom_EQ(tl,ATOMnilptr((b)->ttype),(b)->ttype)?BUNfirst(b):SORTfndfirst_loc(b,tl), \
		  q = atom_EQ(th,ATOMnilptr((b)->ttype),(b)->ttype)?BUNfirst(b):SORTfndlast_loc(b,th); p < q; p++)




#define SORTloop_var(b,p,q,tl,th)					\
	if (!(BATtordered(b)&1)) GDKerror("SORTloop_var: BAT not sorted.\n"); \
	else for (p = atom_EQ(tl,ATOMnilptr((b)->ttype),(b)->ttype)?BUNfirst(b):SORTfndfirst_var(b,tl), \
		  q = atom_EQ(th,ATOMnilptr((b)->ttype),(b)->ttype)?BUNfirst(b):SORTfndlast_var(b,th); p < q; p++)







/* OIDDEPEND */
#if SIZEOF_OID == SIZEOF_INT
#define SORTfnd_oid(b,v)	SORTfnd_int(b,v)
#define SORTfndfirst_oid(b,v)	SORTfndfirst_int(b,v)
#define SORTfndlast_oid(b,v)	SORTfndlast_int(b,v)
#define SORTloop_oid(b,p,q,tl,th) SORTloop_int(b,p,q,tl,th)
#else
#define SORTfnd_oid(b,v)	SORTfnd_lng(b,v)
#define SORTfndfirst_oid(b,v)	SORTfndfirst_lng(b,v)
#define SORTfndlast_oid(b,v)	SORTfndlast_lng(b,v)
#define SORTloop_oid(b,p,q,tl,th) SORTloop_lng(b,p,q,tl,th)
#endif
#if SIZEOF_WRD == SIZEOF_INT
#define SORTfnd_wrd(b,v)	SORTfnd_int(b,v)
#define SORTfndfirst_wrd(b,v)	SORTfndfirst_int(b,v)
#define SORTfndlast_wrd(b,v)	SORTfndlast_int(b,v)
#define SORTloop_wrd(b,p,q,tl,th) SORTloop_int(b,p,q,tl,th)
#else
#define SORTfnd_wrd(b,v)	SORTfnd_lng(b,v)
#define SORTfndfirst_wrd(b,v)	SORTfndfirst_lng(b,v)
#define SORTfndlast_wrd(b,v)	SORTfndlast_lng(b,v)
#define SORTloop_wrd(b,p,q,tl,th) SORTloop_lng(b,p,q,tl,th)
#endif
#define SORTloop_bit(b,p,q,tl,th) SORTloop_chr(b,p,q,tl,th)



#define GDK_AGGR_SIZE 1
#define GDK_AGGR_CARD 2
#define GDK_AGGR_HASNIL 3

gdk_export void PROPdestroy(PROPrec *p);
gdk_export bit BATgetprop_bit(BAT *b, int idx);
gdk_export int BATgetprop_int(BAT *b, int idx);
gdk_export wrd BATgetprop_wrd(BAT *b, int idx);
gdk_export const char *BATgetprop_str(BAT *b, int idx);
gdk_export void BATsetprop_bit(BAT *b, int idx, bit val);
gdk_export void BATsetprop_int(BAT *b, int idx, int val);
gdk_export void BATsetprop_wrd(BAT *b, int idx, wrd val);
gdk_export void BATsetprop_str(BAT *b, int idx, str val);
gdk_export void BATpropagate(BAT *dst, BAT *src, int idx);
gdk_export BAT *BATsample(BAT *b, size_t n);
gdk_export BAT *BAThistogram(BAT *b);
gdk_export size_t BATtopN(BAT *b, size_t topN);	/* used in monet5/src/modules/kernel/algebra.mx */





#define AGGR_MAGIC	111
#define NOID(x)		((oid)(x))
#define NOID_AGGR(x)	NOID_MULT(AGGR_MAGIC,x)
#define NOID_MULT(x,y)	NOID( (lng)(y)*(lng)(x) )






#define JOIN_EQ	0
#define JOIN_LT	-1
#define JOIN_LE	-2
#define JOIN_GT	1
#define JOIN_GE	2
#define JOIN_BAND 3

gdk_export BAT *BATfragment(BAT *b, ptr hl, ptr hh, ptr tl, ptr th);
gdk_export BAT *BAT_select_(BAT *b, ptr tl, ptr th, bit li, bit hi, bit tail, bit preserve_order);
gdk_export BAT *BAT_select(BAT *b, ptr tl, ptr th, bit tail);
gdk_export BAT *BATselect_(BAT *b, ptr tl, ptr th, bit li, bit hi);
gdk_export BAT *BATuselect_(BAT *b, ptr tl, ptr th, bit li, bit hi);
gdk_export BAT *BATselect(BAT *b, ptr tl, ptr th);
gdk_export BAT *BATuselect(BAT *b, ptr tl, ptr th);
gdk_export BAT *BATrestrict(BAT *b, ptr hl, ptr hh, ptr tl, ptr th);

gdk_export BAT *BATconst(BAT *l, int tt, ptr val);
gdk_export BAT *BATthetajoin(BAT *l, BAT *r, int mode, size_t estimate);
gdk_export BAT *BATleftthetajoin(BAT *l, BAT *r, int mode, size_t estimate);
gdk_export BAT *BATnlthetajoin(BAT *l, BAT *r, int mode, size_t estimate);
gdk_export BAT *BATsemijoin(BAT *l, BAT *r);
gdk_export BAT *BATmergejoin(BAT *l, BAT *r, size_t estimate);
gdk_export BAT *BATleftmergejoin(BAT *l, BAT *r, size_t estimate);
gdk_export BAT *BATjoin(BAT *l, BAT *r, size_t estimate);
gdk_export BAT *BATantijoin(BAT *l, BAT *r);
gdk_export BAT *BATleftjoin(BAT *l, BAT *r, size_t estimate);
gdk_export BAT *BATouterjoin(BAT *l, BAT *r, size_t estimate);
gdk_export BAT *BATcross(BAT *l, BAT *r);

gdk_export BAT *BATslice(BAT *b, size_t low, size_t high);
gdk_export BAT *BATfetch(BAT *b, BAT *s);
gdk_export BAT *BATfetchjoin(BAT *b, BAT *s, size_t estimate);
gdk_export BAT *BATleftfetchjoin(BAT *b, BAT *s, size_t estimate);

gdk_export BAT *BATins_kunique(BAT *bn, BAT *b);
gdk_export BAT *BATsunique(BAT *b);
gdk_export BAT *BATkunique(BAT *b);
gdk_export BAT *BATukunique(BAT *b);
gdk_export BAT *BATsintersect(BAT *b, BAT *c);
gdk_export BAT *BATkintersect(BAT *b, BAT *c);
gdk_export BAT *BATsunion(BAT *b, BAT *c);
gdk_export BAT *BATkunion(BAT *b, BAT *c);
gdk_export BAT *BATsdiff(BAT *b, BAT *c);
gdk_export BAT *BATkdiff(BAT *b, BAT *c);

/* generic n-ary multijoin beast, with defines to interpret retval */
#define MULTIJOIN_SORTED(r)	((char*) &r)[0]
#define MULTIJOIN_KEY(r)	((char*) &r)[1]
#define MULTIJOIN_SYNCED(r)	((char*) &r)[2]
#define MULTIJOIN_LEAD(r)	((char*) &r)[3]

typedef void (*ColFcn) (ptr, ptr);
typedef void (*RowFcn) (ptr, ptr *);

gdk_export int BATmultijoin(int argc, BAT *argv[], RowFcn tuple_fcn, ptr tuple_data, ColFcn value_fcn[], ptr value_data[], int orderspec);

#ifdef __cplusplus
}
#endif




#define ILLEGALVALUE	((ptr)-1L)
#define MAXPARAMS	32

#endif /* _GDK_H_ */

